/* GTEST: A genetic algorithm implementation to test sorting
   BY:    Clark Verbrugge
   DATE:  February, 1992
          (modified October, 1992)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#define MAXBUFF 80          /* Input buffer size when reading parameters */

#define LENGTH 13           /* Length of a string */
#define BLANK -1            /* Value of a blank if decoded */
#define RANGE 100           /* Values in template */
#define SPLICEPROB 600      /* Prob out of 1000 of splicing */
#define MUTATEPROB 1        /* Prob out of 1000 of mutation */
#define FMULTIPLE 2.0       /* For fitness scaling */
#define DEFAULTLOOPS 100    /* Default number of generations */
#define DEFAULTSEED 917     /* Default random seed */
#define DEFAULTPOPSIZE 200  /* Default population size */
#define FITFUNCS 3          /* Number of fitness functions */

/* Structure to allow easy read of input file */
typedef struct param_type {
  char *s;
  int *p;
} params;

/* Different fitness functions used */
char *ffs1[FITFUNCS] = {
    "Manhattan distance",
    "Linear increase",
    "Pair-ordered"
};

int *p;                     /* Population of encoded solutions */
double *f;                  /* Fitness values for above */
int *pool;                  /* Mating pool */
int d[LENGTH];              /* Global decoded string */
int t[LENGTH];              /* Template for decoding */
int ts[LENGTH];             /* Template for decoding, sorted */

int best[LENGTH];           /* The best string so far */
double bestf;               /* Raw fitness value of best */

int popsize = DEFAULTPOPSIZE;   /* Size of population */

double sum;                 /* Stat variables in both double */
double max;                 /* (scaled) format,... */
double min;
int intsum;                 /* ..and integer (unscaled) format */
int intmax;
int intmin;
int mutates;                /* Number of mutations in last gen */
double scalea;              /* Fitness scaling coefficients */
double scaleb;

int scaling = 1;            /* Fitness scaling on by default */
int fitfuncchoice;          /* Fitness function to use */
int loops;                  /* Generations to run through */
int popsize;                /* Population size */
int rseed;                  /* Random number generator seed */


params paramlist[] = {
  {"scaling",&scaling},
  {"generations",&loops},
  {"seed",&rseed},
  {"popsize",&popsize},
  {"fitfunc",&fitfuncchoice}
};

int randomnum(int n);
void create(void);
int intcompare(int *i1,int *i2);
int printpop(int gen);
int printstr(int *s);
void copystr(int *pfrom,int *pto);
double fitfunc(int *s);
int fitness(void);
void prescale(void);
int findnext(int curp,int offs);
int decode(int *s);
void matingpool(void);
void nextpop(void);
void splice(int *from1,int *from2,int *to1,int *to2,int pos);
void copybit(int *sfrom,int *sto,int pos);
int preallocate(void);
void deallocate(void);

/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/
int main(void)
    {
    int i,j;
	char buff[MAXBUFF+1];
	 
    loops=DEFAULTLOOPS;
    rseed=DEFAULTSEED;

	bestf= -1e32;

	/* Read in parameters from input */
	while(gets(buff)!=NULL) {

	  for(j=0;j<sizeof(paramlist)/sizeof(params);j++) {

		if(!strncmp(buff,paramlist[j].s,strlen(paramlist[j].s))) {

		  *(paramlist[j].p) = atoi(buff+strlen(paramlist[j].s)+1);
		  printf("--> %s: %d\n",paramlist[j].s,*(paramlist[j].p));
		  break;
		}
		  
	  }
	}

    if(preallocate()) {
        fprintf(stderr,"\n**ERROR: Unable to allocate memory");
        deallocate();
        exit(1);
    }

    printf("\nPopulation size: %d",popsize);
    printf("\nNumber of generations: %d",loops);
    printf("\nUsing random seed: %d",rseed);
    printf("\nFitness function: %s",ffs1[fitfuncchoice]);
    printf("\nScaling: %s\n",(scaling ? "ON" : "OFF"));

    srand(rseed);

    /* First generate template (string of numbers to sort) */
    for(i=0;i<LENGTH;i++)
        t[i]=randomnum(RANGE);
    printf("\nTemplate: ");
    printstr(t);

    copystr(t,ts);
    qsort(ts,LENGTH,sizeof(int),intcompare);
    printf("\n(sorted): ");
    printstr(ts);
	printf("\n");
    
    /* Create the initial population */
    create();

	/* Population printing has been removed to make the output a little
	   more manageable 
    printf("\nInitial population");
    printpop(0);
	*/

	printf("\n(Evolving--please wait)\n");

    for(i=0;i<loops;i++) {
        matingpool();
        nextpop();
        fitness();
        /*printpop(i+1);*/
    }

	printf("\n  Result: ");
	decode(best);
	printstr(d);
	printf("\n fitness: %g\n",bestf);

    deallocate();

    return(0);
}

/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/
int randomnum(int n)
    /* Returns a random number between 0 and n-1 inclusive */
    {
    return(rand()%n);
}

/*------------------------------------------------------------------*/
void create(void) {
    /* Generates the initial population */
    int i,j;

    for(i=0;i<popsize;i++) {
        for(j=0;j<LENGTH;j++) {
            p[i*LENGTH+j]=randomnum(LENGTH-j);
        }
    }

    fitness();
}

/*------------------------------------------------------------------*/
int intcompare(int *i1,int *i2)
    /* Comparison routine used by qsort */
    {
    return((*i1 > *i2) ? 1 : ( *i1 == *i2 ? 0 : -1));
}

/*------------------------------------------------------------------*/
int printpop(int gen)
    /* Prints out the entire current population, along with
       decoded values and fitness values */
    {
    int i,cnt;

    printf("\n\nGeneration: %d",gen);
    printf("\nMinimum: %d (%g)\nAverage: %d (%g)\nMaximum: %d (%g)",
        intmin,min,
        intsum/popsize,sum/popsize,
        intmax,max);
    printf("\nMutates: %d\n",mutates);

    for(i=0;i<LENGTH*3;i++)
        putchar('-');
    printf("    ");
    for(i=0;i<LENGTH*3;i++)
        putchar('-');
    printf("  -----\n");

    for(i=0,cnt=0;i<popsize;i++) {
        if(f[i]<max) continue;
        if(cnt>0) {
            cnt++;
            continue;
        }
        cnt=1;
        printstr(&(p[i*LENGTH]));
        decode(&(p[i*LENGTH]));
        printf(" -> ");
        printstr(d);
        printf("%5g\n",f[i]);
    }
    if(cnt>1) printf("[x%d = %d%%]\n",cnt,(cnt*100)/popsize);


    return(0);
}

/*------------------------------------------------------------------*/
int printstr(int *s)
    /* Prints a single string */
    {
    int i;

    for(i=0;i<LENGTH;i++) {
        printf("%2d ",s[i]);
    }

    printf(" ");

    return(0);
}


/*------------------------------------------------------------------*/
void copystr(int *pfrom,int *pto)
    /* Copies a string */
    {
    int i;

    for(i=0;i<LENGTH;i++)
        pto[i] = pfrom[i];
}

/*------------------------------------------------------------------*/
double fitfunc(int *s)
    /* Computes the fitness coefficient of the given string */
    {
    double ff;
    int i,j;

    ff=0.0;


    if(fitfuncchoice==0) { 
        /* This fitfunc uses a sort of 'Manhattan distance'
           function to determine fitness */
    
        for(i=0;i<LENGTH;i++) {
            for(j=0;j<LENGTH;j++)
                if(s[i]==ts[j]) break;
            ff += (double)(abs(j-i));
        }

        ff = 100.0-ff;

    } else if(fitfuncchoice==1) {

        for(i=0;i<LENGTH;i++)
            for(j=0;j<i;j++)
                if(s[j] < s[i]) ff++;
    } else if(fitfuncchoice==2) {

        for(i=0;i<LENGTH-1;i++)
            if(s[i] < s[i+1]) ff++;
    }

    return(ff);
}

/*------------------------------------------------------------------*/
int fitness(void)
    /* Calculates the fitness value of every string
       and fills out the f[] array, sum, min and max
       stat variables. If scaling is on also scales all
       fitness values. */
    {
    int i,scaledfit;

    min=LENGTH*LENGTH;

    for(sum=0.0,max=0.0,i=0;i<popsize;i++) {

        decode(&(p[i*LENGTH]));

        f[i]=fitfunc(d);

        if(f[i]>max) max = f[i];
        if(f[i]<min) min = f[i];

		/* Keep best string found in any generation */
		if(max>bestf) {
		  bestf=max;
		  copystr(&(p[i*LENGTH]),best);
		}

        sum += f[i];
    }

    intsum=(int)sum;
    intmax=(int)max;
    intmin=(int)min;

    if(scaling) {
    /* Raw fitness scores have now been calculated for the pop
       Now to scale all the fitness scores */

        prescale();

        min=1e30;

        for(max=0.0,sum=0.0,i=0;i<popsize;i++) {
            scaledfit = scalea*f[i]+scaleb;
			f[i]=scaledfit;
            if(f[i]>max) max=f[i];
            if(f[i]<min) min = f[i];
            sum += f[i];
        }
    }

    return(0);
}

/*------------------------------------------------------------------*/
void prescale(void)
    {
    double uavg,delta;

    uavg=sum/popsize;

    if(min>(FMULTIPLE*uavg - max)/(FMULTIPLE - 1.0)) {
        delta = max-uavg;
        scalea = (FMULTIPLE - 1.0) * uavg / delta;
        scaleb = uavg*(max - FMULTIPLE*uavg)/delta;
    } else {
        delta = uavg - min;
        scalea = uavg/delta;
        scaleb = -min*uavg/delta;
    }
}

/*------------------------------------------------------------------*/
int findnext(int curp,int offs)
    /* Locates the offs-th next BLANK entry in d from curp */
    {

    while(d[curp]!=BLANK)
        curp=(curp+1)%LENGTH;

    while(offs>0) {
        curp=(curp+1)%LENGTH;
        while(d[curp]!=BLANK)
            curp=(curp+1)%LENGTH;
        offs--;
    }

    return(curp);
}

/*------------------------------------------------------------------*/
int decode(int *s)
    /* Decodes the given string into d */
    {
    int i,curp;

    for(i=0;i<LENGTH;i++)
        d[i] = BLANK;
    
    curp=0;
    for(i=0;i<LENGTH;i++) {
        curp=findnext(curp,s[i]);
        d[curp]=t[i];
    }

    return(0);
}

/*------------------------------------------------------------------*/
void matingpool(void) {
    /* Selects the mating pool */
    int i,j;
    double pr,r;

    /* Select a population of mates */
    for(i=0;i<popsize;i++) {
        r = randomnum((int)sum);
        pr=f[0];
        j=0;
        while(pr<r) {
            j++;
            pr += f[j];
        }
        copystr(&(p[j*LENGTH]),&(pool[i*LENGTH]));
    }
}

/*------------------------------------------------------------------*/
void nextpop(void) 
    /* Generates the next population from the mating pool */
    {
    int i,splicept;

    mutates = 0;

    for(i=0;i<popsize;i+=2) {
        if(randomnum(1000)<SPLICEPROB)
            splicept = randomnum(LENGTH-1)+1;
        else
            splicept = 0;
        splice(&(pool[i*LENGTH]),&(pool[(i+1)*LENGTH]),
            &(p[i*LENGTH]),&(p[(i+1)*LENGTH]),splicept);
    }
}

/*------------------------------------------------------------------*/
void splice(int *from1,int *from2,int *to1,int *to2,int pos)
    {
    int i;

    for(i=0;i<pos;i++) {
        copybit(from1,to1,i);
        copybit(from2,to2,i);
    }

    for(;i<LENGTH;i++) {
        copybit(from2,to1,i);
        copybit(from1,to2,i);
    }
}

/*------------------------------------------------------------------*/
void copybit(int *sfrom,int *sto,int pos)
    {

    if(randomnum(1000)<MUTATEPROB) {
        sto[pos] = randomnum(LENGTH-pos);
        mutates++;
    } else
        sto[pos] = sfrom[pos];
}

int preallocate(void)
    /* Builds dynamic arrays according to current population size */
    {

    p=NULL;
    pool=NULL;
    f=NULL;

    p=(int *)malloc(sizeof(int)*LENGTH*popsize);
    if(p==NULL) return(-1);

    pool=(int *)malloc(sizeof(int)*LENGTH*popsize);
    if(pool==NULL) return(-1);

    f=(double *)malloc(sizeof(double)*popsize);
    if(f==NULL) return(-1);

    return(0);
}
void deallocate(void)
    /* Deallocates allocated memory */
    {

    if(p!=NULL) free(p);
    if(pool!=NULL) free(pool);
    if(f!=NULL) free(f);

    p=NULL;
    pool=NULL;
    f=NULL;
}



