/* Program to convert a file from binary format to ASCII dinero format */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>

#define XREAD 0
typedef unsigned long ADDRESSTYPE;

typedef struct {
  ADDRESSTYPE address;
} trace_record;

#ifndef N_BUFS
#define N_BUFS 	4
#endif

#define OK		0
#define ERR		-1
#define ILLEGALNUM	-29284

static void echocmdargs();
static int errormessage();	
static int atoiKMG();

/* Macro to extract the argument for a option . Returns pointer to it. */
/* Increments i (the index into argv) if needed */
#define ARGVAL(i)	(argv[i][2]=='\0' ? argv[++i] : &argv[i][2])

/* Analyze command line arguments. 
** The options recognized by tycho are :
** -z<num>	: max ref. count	(default - INT_MAX)
** -f<num>	: Buffer space(in pages)(default - 4)
*/ 

int getcmdargs(register int argc,register char *argv[],int *nrefp,
               int *bufsize)
{ 
register int i;
register int nextnum;
int  returnflag;

returnflag = OK;

/* defaults */
*nrefp = INT_MAX;
/*
*bufsize = N_BUFS * getpagesize();;
*/

*bufsize = N_BUFS * sysconf(_SC_PAGESIZE);

/* 
**	Echo command line 
*/
echocmdargs(argc,argv);

/* 
**	Process arguments 
*/
for (i = 1; i < argc; i++) 
if (argv[i][0] != '-') {
          printf("Argument %d (%s) must be preceded by '-'.\n",i,argv[i]);
	  returnflag = ERR;
}
else
{
	switch (argv[i][1])	/* switch on second char. */
	{			/* allow: "-m100" or "-m 100" */

/*
**	CACHE
*/

case 'z' :
	nextnum = atoiKMG(ARGVAL(i));
	if (nextnum >= 0 )
		*nrefp = nextnum;
	else {
		*nrefp = ILLEGALNUM;
		returnflag = ERR;
	}
	break;

case 'f' :
	nextnum = atoiKMG(ARGVAL(i));
	if (nextnum > 0 )
		*bufsize = nextnum * sysconf(_SC_PAGESIZE);
	else {
		*bufsize = ILLEGALNUM;
		returnflag = ERR;
	}
	break;

default :
	printf("Argument %d is invalid: %s\n",i,argv[i]);
	returnflag = ERR;
	break;
 	} 
}

/* Check parameters for correctness */

returnflag = errormessage(returnflag,(returnflag==ERR),
				"Illegal parameter or value.");

return(returnflag);
} /* ************************************************************ */


/*
** echocmdargs : Echoes the command line arguments to stdout.
*/
static void echocmdargs(int argc,char *argv[])
{
register int i;

	fprintf(stderr,"\nCmdline is ");

	/* Print args separated by blanks */
	for (i = 0; i < (argc - 1); i++)
		fprintf(stderr,"%s ", argv[i]);

	/* Print last arg. and newline */
	fprintf(stderr,"%s\n", argv[i]);
}



/* Prints error message */
static int errormessage(int olderrorflag,int condition,char *errmessage)
{
register int newerrorflag;

/*
** E.G. USE:
**
**	returnflag = errormessage(returnflag,
**	    (cachep->Isize==ILLEGALNUM),
**	    "Instruction cache size not specified."
**	);
*/

newerrorflag = olderrorflag;

if (condition) {
	printf("ERROR: %s\n\n", errmessage);
	newerrorflag = ERR;
 }

return(newerrorflag);
} /* ************************************************************ */
static int atoiKMG(char *str)
/*
**  Return integer equal to value of ascii string that is sequence
**  of digits possibly followed by one of the following suffixes:
**
**	K, k	1024
**	M, m	1024**2 = 1048576
**	G, g	1024**3 = 1073741824
**
**  E.g., "64K" evaluated to integer 65536.
**  ILLEGALNUM is returned for illegal suffexes or on overflow.
*/
{
char str_buffer[256];
int length;
char suffix;
int root, multiplier, value;

strcpy(str_buffer,str);

length = strlen(str_buffer);
suffix = str_buffer[length-1];

switch (suffix) {

case 'K' :
case 'k' :
	multiplier = 1024;
	str_buffer[length-1] = NULL;
	break;

case 'M' :
case 'm' :
	multiplier = 1048576;		/* 1024**2 */
	str_buffer[length-1] = NULL;
	break;

case 'G' :
case 'g' :
	multiplier = 1073741824;	/* 1024**3 */
	str_buffer[length-1] = NULL;
	break;

case '0' :
case '1' :
case '2' :
case '3' :
case '4' :
case '5' :
case '6' :
case '7' :
case '8' :
case '9' :
	multiplier = 1;
	break;

default :
	multiplier = ILLEGALNUM;
	break;
 
 }


if (multiplier!=ILLEGALNUM) {
	root = atoi(str_buffer);
	value = root*multiplier;

	if (root!=(value/multiplier)) {
		value = ILLEGALNUM;
	}
 }
else {
	value = ILLEGALNUM;
 }


return(value);
} /* ***************************************************************** */
void main(int argc,char *argv[])
{
    int i;
    int nrefs;
    int buffersize;
    int refs_written = 0, bytes_read;
    trace_record *buffer;
    unsigned long refs_per_block;

    if (getcmdargs(argc,argv,&nrefs,&buffersize) != OK) {
	printf("Error in command line arguments\n");
	printf("Usage %s -z max.ref.count -f buffersize(in pages)",argv[0]);
	exit(-1);
    }

    refs_per_block = (buffersize / sizeof(trace_record));
    buffer = (trace_record *) malloc(refs_per_block * sizeof(trace_record));
    while (nrefs) {
	/* If last block , reduce refs_per_block */
	if (nrefs < refs_per_block)
		refs_per_block = nrefs;

	bytes_read = read(0,(char *) buffer,sizeof(trace_record) * refs_per_block);
	if ((bytes_read == EOF) || (bytes_read == 0))
		break;
	if ((bytes_read < 0) || (bytes_read % sizeof(trace_record))) {
		fprintf(stderr,"\n **** Error in standard input.\n");
    		fprintf(stderr,"Wrote out %d trace records\n",refs_written);
		exit(-1);
	}

	if (refs_per_block > bytes_read/sizeof(trace_record)) 
		refs_per_block = bytes_read/sizeof(trace_record);

	for (i = 0; i < refs_per_block; i++) {
	    printf("%x %x\n",XREAD,buffer[i].address);
	}
	nrefs -= refs_per_block;
	refs_written += refs_per_block;
    }
    fprintf(stderr,"Wrote out %d trace records\n",refs_written);
    exit(0);
}
