/* Program to convert a file in ASCII dinero format to binary format */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>

typedef unsigned long LABELTYPE;
typedef unsigned long ADDRESSTYPE;

typedef struct {
  ADDRESSTYPE address;
} trace_record;

#ifndef N_BUFS
#define N_BUFS 	4
#endif

#define OK		0
#define ERR		-1
#define ILLEGALNUM	-29284

static void Calculate_HexValue_Table();
static void echocmdargs();
static int sscanxx();
static int errormessage();	
static int atoiKMG();

/* To sscanxx */

#define HEX	4

#define WHITE_SPACE	((*ptr==' ')||(*ptr=='\t'))

#define CHAR_DIGIT_OFFSET		('0')
#define CHAR_LOWERCASE_ABCDEF_OFFSET	('a' - 10)
#define CHAR_UPPERCASE_ABCDEF_OFFSET	('A' - 10)

/* table to do translation from character to hex digits for sscanxx */
static unsigned char HexCharValue[256];

#define CHAR_HEX_DIGIT  ((digit = HexCharValue[*ptr]) < 16)

/* Values returned by HexCharValue for : */
#define WHITESPACECHAR	253	/* For field separators */
#define XCHAR		254	/* characters 'x' and 'X' */
#define ILLEGALCHAR	255	/* Illegal characters (excluding white space)*/

static void Calculate_HexValue_Table(table)
unsigned char *table;
{
register unsigned int i;

	/* initialize to illegal */
	for (i = 0; i < 256; i++) 
		table[i] = ILLEGALCHAR;

	/* '0' to '9' */
	for (i = '0'; i <= '9'; i++)  
		table[i] = i - CHAR_DIGIT_OFFSET;

	/* for 'a' to 'f' */
	for (i = 'a'; i <= 'f'; i++)  
		table[i] = i - CHAR_LOWERCASE_ABCDEF_OFFSET;

	/* for 'A' to 'F' */
	for (i = 'A'; i <= 'F'; i++)  
		table[i] = i - CHAR_UPPERCASE_ABCDEF_OFFSET;

	/* 'x' and 'X' */
	table['x'] = table['X'] = XCHAR;

	/* white space, field seperators */
	table[' '] = table['\t'] = WHITESPACECHAR;
}

static int sscanxx(char linebuffer[],unsigned long *num1p,unsigned long *num2p)
/*
**	This routine is designed to do the same function as
**	sscanf(linebuffer,"%x %x",labelp,addrp).  It was added
**	because gprof said that 35 to 50% of dineroIII's run time
**	was being spent in sscanf.   This routine runs about 10	times faster 
**	than sscanf and >20% faster than the old version of tycho.
**
**	Note: Use of 0x prefix causes slower execution as parsing takes longer.
**
**	The routine is not exactly the same as sscanf:
**	
**	1) It differs in the way it handles the 0x (or 0X) prefix in parsing
**	hex numbers in the input. It accepts numbers 
**	a) which are just 0x or 0X to be equal to 0.
**	b) 0x0x0x{...} or 0X0X0X{...} or any combination of such prefixes
**	as valid. So the following numbers are accepted :
**	0x0x38Fbf 0X0x348fbe 0X0X0X0x0X0x8Fbe  ....
**	    This compromise was made as the decrease in speed to catch this 
**	rare error was not considered worth it.
**
**	This routine adds the following features to the old version of tycho :
**	1) Does Error checking.
**	2) Recognises 0x or 0X prefix before hex numbers.
**	3) Recognises 'A'-'F' too as valid hex numbers.
**	But 4) It needs a table to be initialized.
*/
{
register unsigned char digit;
register char *ptr;
register unsigned long num;

/*
**	Exit with EOF if the buffer is empty.
*/
if (*(ptr = linebuffer)==NULL) return(EOF);

while (WHITE_SPACE) ptr++;

/*
**	Convert first hex number.
*/
while (1) {
	num = 0;
	while (CHAR_HEX_DIGIT) {
		num = (num << HEX) + digit;
		ptr++;
 	}
	if (digit == WHITESPACECHAR) break;	/* End of word */
	else if (digit == XCHAR) {
		if (num != 0) return(0);
		else ptr++;		/* If 0X or 0x restart, else error */
	}
	else if (digit == ILLEGALCHAR) return(0);    /* Illegal char in input */
}

*num1p = num;

/*	Skip white space.	*/
while (WHITE_SPACE) ptr++;

/*
**	Convert second hex number.
*/
while (1) {
	num = 0;
	while (CHAR_HEX_DIGIT) {
		num = (num << HEX) + digit;
		ptr++;
 	}
	if ((*ptr == '\0') || (digit == WHITESPACECHAR)) break;/* End of word */
	else if (digit == XCHAR) {
		if (num != 0) return(0);
		else ptr++;		/* If 0X or 0x restart, else error */
	}
	else if (digit == ILLEGALCHAR) return(1);    /* Illegal char in input */
}

*num2p = num;

return(2);
}

int readfrominputstream(LABELTYPE *labelp,ADDRESSTYPE *addrp) 
                          /* get next addr */
{
register int readflag;
char linebuffer[255];
char *gets();	/* Gets chars from stdin until a newline;
		** changes newline to nullchar; returns ptr
		** to the sting.
		*/

/* 
**	Read in ASCII from standard input 
**	Expect ONE label and addr in hex per line.
**	Rest of data input line is ignored so it may
**	be used for comments.
**
**	WARNING: If more than one tuple is put on a line,
**	all but the first tuple will be ignored.
*/

linebuffer[0] = NULL;
gets(linebuffer);

readflag = sscanxx(linebuffer,labelp,addrp);

if (readflag!=2)
{
	if (readflag!=EOF) {
		fprintf(stderr,"\n **** Error in standard input.\n");
		exit(-1);
	}
	return(EOF);
}

} /* ********************************************************************* */

/* Macro to extract the argument for a option . Returns pointer to it. */
/* Increments i (the index into argv) if needed */
#define ARGVAL(i)	(argv[i][2]=='\0' ? argv[++i] : &argv[i][2])

/* Analyze command line arguments. 
** The options recognized by tycho are :
** -z<num>	: max ref. count	(default - INT_MAX)
** -f<num>	: Buffer space(in pages)(default - 4)
*/ 

int getcmdargs(register int argc,register char *argv[],int *nrefp,
               int *bufsize)
{ 
register int i;
register int nextnum;
int  returnflag;

returnflag = OK;

/* defaults */
*nrefp = INT_MAX;
*bufsize = N_BUFS * sysconf(_SC_PAGESIZE);

/* 
**	Echo command line 
*/
echocmdargs(argc,argv);

/* 
**	Process arguments 
*/
for (i = 1; i < argc; i++) 
if (argv[i][0] != '-') {
          printf("Argument %d (%s) must be preceded by '-'.\n",i,argv[i]);
	  returnflag = ERR;
}
else
{
	switch (argv[i][1])	/* switch on second char. */
	{			/* allow: "-m100" or "-m 100" */

/*
**	CACHE
*/

case 'z' :
	nextnum = atoiKMG(ARGVAL(i));
	if (nextnum >= 0 )
		*nrefp = nextnum;
	else {
		*nrefp = ILLEGALNUM;
		returnflag = ERR;
	}
	break;

case 'f' :
	nextnum = atoiKMG(ARGVAL(i));
	if (nextnum > 0 )
		*bufsize = nextnum * sysconf(_SC_PAGESIZE);
	else {
		*bufsize = ILLEGALNUM;
		returnflag = ERR;
	}
	break;

default :
	printf("Argument %d is invalid: %s\n",i,argv[i]);
	returnflag = ERR;
	break;
 	} 
}

/* Check parameters for correctness */

returnflag = errormessage(returnflag,(returnflag==ERR),
				"Illegal parameter or value.");

return(returnflag);
} /* ************************************************************ */


/*
** echocmdargs : Echoes the command line arguments to stdout.
*/
static void echocmdargs(int argc,char *argv[])
{
register int i;

	fprintf(stderr,"\nCmdline is ");

	/* Print args separated by blanks */
	for (i = 0; i < (argc - 1); i++)
		fprintf(stderr,"%s ", argv[i]);

	/* Print last arg. and newline */
	fprintf(stderr,"%s\n", argv[i]);
}



/* Prints error message */
static int errormessage(int olderrorflag,int condition,char *errmessage)
{
register int newerrorflag;

/*
** E.G. USE:
**
**	returnflag = errormessage(returnflag,
**	    (cachep->Isize==ILLEGALNUM),
**	    "Instruction cache size not specified."
**	);
*/

newerrorflag = olderrorflag;

if (condition) {
	printf("ERROR: %s\n\n", errmessage);
	newerrorflag = ERR;
 }

return(newerrorflag);
} /* ************************************************************ */


static int atoiKMG(char *str)
/*
**  Return integer equal to value of ascii string that is sequence
**  of digits possibly followed by one of the following suffixes:
**
**	K, k	1024
**	M, m	1024**2 = 1048576
**	G, g	1024**3 = 1073741824
**
**  E.g., "64K" evaluated to integer 65536.
**  ILLEGALNUM is returned for illegal suffexes or on overflow.
*/
{
char str_buffer[256];
int length;
char suffix;
int root, multiplier, value;

strcpy(str_buffer,str);

length = strlen(str_buffer);
suffix = str_buffer[length-1];

switch (suffix) {

case 'K' :
case 'k' :
	multiplier = 1024;
	str_buffer[length-1] = NULL;
	break;

case 'M' :
case 'm' :
	multiplier = 1048576;		/* 1024**2 */
	str_buffer[length-1] = NULL;
	break;

case 'G' :
case 'g' :
	multiplier = 1073741824;	/* 1024**3 */
	str_buffer[length-1] = NULL;
	break;

case '0' :
case '1' :
case '2' :
case '3' :
case '4' :
case '5' :
case '6' :
case '7' :
case '8' :
case '9' :
	multiplier = 1;
	break;

default :
	multiplier = ILLEGALNUM;
	break;
 
 }


if (multiplier!=ILLEGALNUM) {
	root = atoi(str_buffer);
	value = root*multiplier;

	if (root!=(value/multiplier)) {
		value = ILLEGALNUM;
	}
 }
else {
	value = ILLEGALNUM;
 }


return(value);
} /* ***************************************************************** */

void main(int argc,char *argv[])
{
    int i;
    int nrefs;
    int buffersize;
    int refs_written = 0;
    trace_record *buffer;
    unsigned long refs_per_block;
    LABELTYPE  dummy;

    if (getcmdargs(argc,argv,&nrefs,&buffersize) != OK) {
	printf("Error in command line arguments\n");
	printf("Usage %s -z max.ref.count -f buffersize(in pages)",argv[0]);
	exit(-1);
    }

    Calculate_HexValue_Table(HexCharValue);

    refs_per_block = (buffersize / sizeof(trace_record));
    buffer = (trace_record *) malloc(refs_per_block * sizeof(trace_record));
    while (nrefs) {
	/* If last block , reduce refs_per_block */
	if (nrefs < refs_per_block)
		refs_per_block = nrefs;

	for (i = 0; i < refs_per_block; i++) {
	    if(readfrominputstream(&dummy,&(buffer[i].address)) == EOF) {
		if (i > 0)
			write(1,(char *) buffer,sizeof(trace_record) * i);
		refs_written += i;
		fprintf(stderr,"Wrote out %d trace records\n",refs_written);
		exit(0);
	    }
	}
	write(1,(char *) buffer,sizeof(trace_record) * refs_per_block);
	nrefs -= refs_per_block;
	refs_written += refs_per_block;
    }
    fprintf(stderr,"Wrote out %d trace records\n",refs_written);
    exit(0);
}
